/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

"use strict";

const Cu = Components.utils;
const Cc = Components.classes;
const Cr = Components.results;
const Ci = Components.interfaces;

const SCHEME = "moz-thumb";
const CONTENT_TYPE = "image/png";

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

XPCOMUtils.defineLazyServiceGetter(this, "gAppShellService",
                                   "@mozilla.org/appshell/appShellService;1",
                                   "nsIAppShellService");

XPCOMUtils.defineLazyServiceGetter(this, "gCacheService",
                                   "@mozilla.org/network/cache-service;1",
                                   "nsICacheService");

XPCOMUtils.defineLazyGetter(this, "gNetUtil", function () {
  var obj = {};
  Cu.import("resource://gre/modules/NetUtil.jsm", obj);
  return obj.NetUtil;
});

// TODO
function debug(aMsg) {
  aMsg = ("Thumbnails: " + aMsg).replace(/\S{80}/g, "$&\n");
  Services.console.logStringMessage(aMsg);
}

// TODO
function trace(aMsg) {
  // cut off the first line of the stack trace, because that's just this function.
  let stack = Error().stack.split("\n").slice(1);

  debug("trace: " + aMsg + "\n" + stack.join("\n"));
}

// TODO
function extend(aParent, aObject) {
  function copy(aSource, aDest) {
    for (let name in aSource) {
      let getter, setter;

      if (getter = aSource.__lookupGetter__(name))
        aDest.__defineGetter__(name, getter);
      else if (setter = aSource.__lookupSetter__(name))
        aDest.__defineSetter__(name, setter);
      else
       aDest[name] = aSource[name];
    }

    return aDest;
  }

  return copy(aObject, copy(aParent, {}));
}

// TODO
function URLParser(aUri) {
  this._uri = aUri;
}

URLParser.prototype = {
  parse: function UP_parse() {
    if (this._parsed)
      return this._parsed;

    return this._parsed = this._parseParams();
  },

  _parseParams: function UP_parseParams() {
    let params = {};
    let query = this._uri.spec.split("?")[1] || "";

    query.split("&").forEach(function (param) {
      let [key, value] = param.split("=").map(decodeURIComponent);
      params[key.toLowerCase()] = value;
    });

    return params;
  }
};

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const ABOUT_BLANK_URI = "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAEAAAA" +
                        "BCAYAAAAfFcSJAAAADUlEQVQImWP4////fwAJ+wP9CNHoHgAAAABJ" +
                        "RU5ErkJggg==";

// TODO
function ThumbnailProtocol() {
}

ThumbnailProtocol.prototype = {
  scheme: SCHEME,
  defaultPort: -1,
  protocolFlags: //Ci.nsIProtocolHandler.URI_DANGEROUS_TO_LOAD |
                 Ci.nsIProtocolHandler.URI_LOADABLE_BY_ANYONE |
                 Ci.nsIProtocolHandler.URI_NORELATIVE |
                 Ci.nsIProtocolHandler.URI_NOAUTH,

  newURI: function TP_newURI(aSpec, aOriginCharset, aBaseUri) {
    let uri = Cc["@mozilla.org/network/simple-uri;1"].createInstance(Ci.nsIURI);
    uri.spec = aSpec;
    return uri;
  },

  newChannel: function TP_newChannel(aUri) {
    let {url} = new URLParser(aUri).parse();

    // don't try to load thumbnails for about:blank pages, just deliver a
    // white 1x1 image. so we make sure that tests can use about:blank
    // and don't have to wait until all the images have been rendered.
    if (0 == url.indexOf("about:blank"))
      return Services.io.newChannel(ABOUT_BLANK_URI, "UTF8", null);

    return new RedirectChannel(aUri);
  },

  allowPort: function TP_allowPort(aPort, aScheme) {
    return false;
  },

  classID: Components.ID("{5a4ae9b5-f475-48ae-9dce-0b4c1d347884}"),
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIProtocolHandler])
};

let NSGetFactory = XPCOMUtils.generateNSGetFactory([ThumbnailProtocol]);
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

let BaseChannel = {
  _wasOpened: false,
  _wasStarted: false,
  _isCanceled: false,

  _setup: function BCH__setup(aUri) {
    let {url} = new URLParser(aUri).parse();

    // prevent recursion if a thumbnail for a thumbnail is requested
    if (0 == url.indexOf(SCHEME + "://"))
      aUri = Services.io.newURI(url, "UTF8", null);

    this._uri = aUri;

    // nsIChannel
    this.originalURI = aUri;

    // nsIHttpChannel
    this._requestHeaders = {};
    this._responseHeaders = {"content-type": CONTENT_TYPE};
  },

  _startRequest: function CCH__startRequest() {
    try {
      this._listener.onStartRequest(this, this._context);
    } catch (e) {
      // the listener might throw if the request has been canceled
      this.cancel(Cr.NS_BINDING_ABORTED);
    }

    return !this.canceled;
  },

  _serveData: function CCH__serveData(aDataStream) {
    try {
      let available = aDataStream.available();
      this._listener.onDataAvailable(this, this._context, aDataStream, 0, available);
    } catch (e) {
      // the listener might throw if the request has been canceled
      this.cancel(Cr.NS_BINDING_ABORTED);
    }

    return !this.canceled;
  },

  _stopRequest: function CCH__stopRequest() {
    try {
      this._listener.onStopRequest(this, this._context, this.status);
    } catch (e) {
      // this might throw but is generally ignored
    }

    this._cleanup();
  },

  _addToLoadGroup: function CCH__addToLoadGroup() {
    if (this.loadGroup)
      this.loadGroup.addRequest(this, this._context);
  },

  _removeFromLoadGroup: function CCH__removeFromLoadGroup() {
    if (!this.loadGroup)
      return;

    try {
      this.loadGroup.removeRequest(this, this._context, this.status);
    } catch (e) {
      // this might throw and is ignored
    }
  },

  _cleanup: function CCH__cleanup() {
    this._removeFromLoadGroup();
    this.loadGroup = null;

    this._isPending = false;

    delete this._listener;
    delete this._context;
    delete this._entry;
  },

  // nsIChannel
  contentType: CONTENT_TYPE,
  contentLength: -1,
  owner: null,
  contentCharset: null,
  notificationCallbacks: null,

  get URI() this._uri,
  get securityInfo() null,

  asyncOpen: function BCH_asyncOpen(aListener, aContext) {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  open: function BCH_open() {
    throw Cr.NS_ERROR_NOT_IMPLEMENTED;
  },

  // nsIHttpChannel
  redirectionLimit: 10,
  requestMethod: "GET",
  allowPipelining: true,
  referrer: null,

  get requestSucceeded() true,
  get responseStatus() 200,
  get responseStatusText() "OK",

  isNoCacheResponse: function () false,
  isNoStoreResponse: function () false,

  getRequestHeader: function BCH_getRequestHeader(aHeader) {
    aHeader = aHeader.toLowerCase();
    let headers = this._requestHeaders;

    if (aHeader in headers)
      return headers[aHeader];

    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  setRequestHeader: function BCH_setRequestHeader(aHeader, aValue, aMerge) {
    if (aMerge)
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    this._requestHeaders[aHeader.toLowerCase()] = aValue;
  },

  visitRequestHeaders: function BCH_visitRequestHeaders(aVisitor) {
    this._visitHeaders(this._requestHeaders, aVisitor);
  },

  getResponseHeader: function BCH_getResponseHeader(aHeader) {
    aHeader = aHeader.toLowerCase();
    let headers = this._responseHeaders;

    if (aHeader in headers)
      return headers[aHeader];

    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  setResponseHeader: function BCH_setResponseHeader(aHeader, aValue, aMerge) {
    if (aMerge)
      throw Cr.NS_ERROR_NOT_IMPLEMENTED;

    this._responseHeaders[aHeader.toLowerCase()] = aValue;
  },

  visitResponseHeaders: function BCH_visitResponseHeaders(aVisitor) {
    this._visitHeaders(this._responseHeaders, aVisitor);
  },

  _visitHeaders: function BCH__visitHeaders(aHeaders, aVisitor) {
    for (let name in aHeaders) {
      let value = aHeaders[name];

      try {
        aVisitor.visitHeader(name, value);
      } catch (e) {
        // the visitor throws to stop the iteration
        return;
      }
    }
  },

  // nsIRequest
  _status: Cr.NS_OK,
  _isPending: false,

  loadFlags: Ci.nsIRequest.LOAD_NORMAL,
  loadGroup: null,

  get name() this._uri.spec,
  get status() this._status,

  isPending: function () this._isPending,
  resume: function () {},
  suspend: function () {},

  cancel: function BCH_cancel(aStatus) {
    if (this.canceled)
      return;

    this._isCanceled = true;
    this._status = aStatus;
  },

  // nsIHttpChannelInternal
  documentURI: null,

  get canceled() this._isCanceled,

  QueryInterface: XPCOMUtils.generateQI([Ci.nsIChannel,
                                         Ci.nsIHttpChannel,
                                         Ci.nsIHttpChannelInternal,
                                         Ci.nsIRequest])
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO
function RedirectChannel(aUri) {
  this._setup(aUri);
}

RedirectChannel.prototype = extend(BaseChannel, {
  asyncOpen: function RCH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    this._fetchEntry();
  },

  _fetchEntry: function RCH__fetchEntry() {
    let self = this;
    let uri = this._uri;

    Cache.getReadEntry(this.name, function (aEntry) {
      let channel;

      if (aEntry)
        // entry found, deliver from cache
        channel = new CacheChannel(uri, aEntry);
      else
        // entry not found, create a thumbnail and deliver it when ready
        channel = new GenerateChannel(uri);

      self._redirect(channel);
    });
  },

  _redirect: function RCH__redirect(aChannel) {
    let sink = this.notificationCallbacks.QueryInterface(Ci.nsIChannelEventSink);
    let flags = Ci.nsIChannelEventSink.REDIRECT_INTERNAL;
    let callback = this._createRedirectCallback(aChannel);

    sink.asyncOnChannelRedirect(this, aChannel, flags, callback);
  },

  _createRedirectCallback: function RCH__createRedirectCallback(aChannel) {
    let self = this;

    return {
      onRedirectVerifyCallback: function (aStatus) {
        if (!Components.isSuccessCode(aStatus))
          throw Cr.NS_ERROR_UNEXPECTED;

        // cancel the current request before opening the new one
        self.cancel(Cr.NS_OK);

        aChannel.loadGroup = self.loadGroup;
        aChannel.asyncOpen(self._listener, self._context);
      },

      QueryInterface: XPCOMUtils.generateQI([Ci.nsIAsyncVerifyRedirectCallback])
    };
  }
});
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO
function CacheChannel(aUri, aEntry) {
  this._entry = aEntry;

  try {
    let contentLength = aEntry.getMetaDataElement("content-length");
    this.setResponseHeader("content-length", contentLength);
  } catch (e) {
    // throws if the meta data element doesn't exist, ignore
  }

  this._setup(aUri);
}

CacheChannel.prototype = extend(BaseChannel, {
  asyncOpen: function CCH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    this._serveEntry();
  },

  cancel: function CCH__cancel(aStatus) {
    if (this.canceled)
      return;

    BaseChannel.cancel.call(this, aStatus);

    this._cleanup();
  },

  _serveEntry: function CCH__serveEntry() {
    let self = this;
    let listener = this._listener;
    let context = this._context;

    this._readEntry(function (aDataStream) {
      self._dataStream = aDataStream;

      if (self._startRequest()) {
        self._addToLoadGroup();

        if (self._serveData(aDataStream))
          self._stopRequest();
      }
    });
  },

  _readEntry: function CCH__readEntry(aCallback) {
    let self = this;

    this._readEntryData(function (aDataStream) {
      if (self.canceled)
        return;

      // failed to get the thumbnail data
      if (!aDataStream) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      aCallback(aDataStream);
    });
  },

  _readEntryData: function CCH__readEntryData(aCallback) {
    let inputStream = this._entry.openInputStream(0);

    gNetUtil.asyncFetch(inputStream, function (aDataStream, aStatus, aRequest) {
      inputStream.close();

      if (!Components.isSuccessCode(aStatus) || !aDataStream.available())
        aDataStream = null;

      aCallback(aDataStream);
    });
  }
});
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO
function GenerateChannel(aUri) {
  this._setup(aUri);
}

GenerateChannel.prototype = extend(BaseChannel, {
  asyncOpen: function GCH_asyncOpen(aListener, aContext) {
    if (this._wasOpened)
      throw Cr.NS_ERROR_ALREADY_OPENED;

    if (this.canceled)
      return;

    this._listener = aListener;
    this._context = aContext;

    this._isPending = true;
    this._wasOpened = true;

    let self = this;

    Cache.getWriteEntry(this.name, function (aEntry) {
      // we might've been canceled while waiting for the entry
      if (self.canceled)
        return;

      // for some reason we don't have a valid entry
      if (!aEntry) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      self._createAndServeThumbnail(aEntry);
    });
  },

  _createAndServeThumbnail: function GCH__createAndServeThumbnail(aEntry) {
    if (!this._startRequest())
      return;

    this._addToLoadGroup();

    let self = this;
    let {url, width, height} = new URLParser(this._uri).parse();

    // create a new thumbnail
    Thumbnailer.create(url, width, height, function (aDataStream, aStatusCode) {
      // failed to get the thumbnail data
      if (!aDataStream || !aDataStream.available()) {
        self.cancel(Cr.NS_ERROR_UNEXPECTED);
        return;
      }

      // store thumbnails in the cache only for 2xx status codes
      if (Math.floor(aStatusCode / 100) == 2)
        self._prepareCacheEntry(aEntry, aDataStream);

      if (self._serveData(aDataStream))
        self._stopRequest();
    });
  },

  _prepareCacheEntry: function GHC__prepareCacheEntry(aEntry, aDataStream) {
    // create a tee listener to simultaneously deliver data to the
    // stream listener and to the cache entry's stream
    this._useWriteEntryTeeListener(aEntry);

    // store content-length header
    let available = "" + aDataStream.available();
    aEntry.setMetaDataElement("content-length", available);

    // store this entry for two days
    let unixTime = Math.floor(Date.now() / 1000);
    aEntry.setExpirationTime(unixTime + (86400 * 2));
  },

  _useWriteEntryTeeListener: function GCH__useWriteEntryTeeListener(aEntry) {
    let tee = Cc["@mozilla.org/network/stream-listener-tee;1"]
              .createInstance(Ci.nsIStreamListenerTee);

    let outputStream = aEntry.openOutputStream(0);
    tee.init(this._listener, outputStream);

    this._listener = tee;
  }
});
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const CACHE_CLIENT_ID = "moz-thumb";

// TODO
let Cache = {
  get _cacheSession() {
    let session = gCacheService.createSession(CACHE_CLIENT_ID,
                                              Ci.nsICache.STORE_ON_DISK, true);

    delete this._cacheSession;
    return this._cacheSession = session;
  },

  getReadEntry: function Cache_getReadEntry(aKey, aCallback) {
    let self = this;

    // try to open the desired cache entry
    this._openCacheEntry(aKey, Ci.nsICache.ACCESS_READ, function (aEntry) {
      if (!self._isValidReadEntry(aEntry))
        aEntry = null;

      aCallback(aEntry);
    });
  },

  getWriteEntry: function Cache_getWriteEntry(aKey, aCallback) {
    // try to open the desired cache entry
    this._openCacheEntry(aKey, Ci.nsICache.ACCESS_WRITE, aCallback);
  },

  _isValidReadEntry: function Cache__isValidReadEntry(aEntry) {
    if (!aEntry)
      return false;

    let inputStream = aEntry.openInputStream(0);

    // check if the entry actually contains data
    return inputStream && inputStream.available();
  },

  _openCacheEntry: function Cache__openCacheEntry(aKey, aAccess, aCallback) {
    function onCacheEntryAvailable(aEntry, aAccessGranted, aStatus) {
      let validAccess = (aAccess == aAccessGranted);
      let validStatus = Components.isSuccessCode(aStatus);

      if (aEntry && !(validAccess && validStatus)) {
        aEntry.close();
        aEntry = null;
      }

      aCallback(aEntry);
    }

    let listener = this._createCacheListener(onCacheEntryAvailable);
    this._cacheSession.asyncOpenCacheEntry(aKey, aAccess, listener);
  },

  _createCacheListener: function Cache__createCacheListener(aCallback) {
    return {
      onCacheEntryAvailable: aCallback,
      QueryInterface: XPCOMUtils.generateQI([Ci.nsICacheListener])
    };
  }
};
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Thumbnails code.
 *
 * The Initial Developer of the Original Code is
 * the Mozilla Foundation.
 * Portions created by the Initial Developer are Copyright (C) 2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Tim Taubert <ttaubert@mozilla.com> (Original Author)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const XUL_NS = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const XUL_FRAME_SRC = "chrome://global/content/mozilla.xhtml";

const THUMB_MIN_WIDTH = 1024;
const THUMB_BG_COLOR = "#fff";

// TODO
let Thumbnailer = {
  create: function TN_create(aUrl, aWidth, aHeight, aCallback) {
    let frameWidth = Math.max(aWidth, THUMB_MIN_WIDTH);
    let frameHeight = frameWidth / (aWidth / aHeight);

    let self = this;

    HiddenFrame.create(aUrl, frameWidth, frameHeight, function (aFrame, aStatusCode) {
      // draw to the canvas as soon as the frame is initialized
      ctx.drawWindow(aFrame.contentWindow, 0, 0, frameWidth, frameHeight,
                     THUMB_BG_COLOR, ctx.DRAWWINDOW_DO_NOT_FLUSH);

      // read the image data from the canvas and pass it to the callback
      self._readCanvasData(canvas, function (aDataStream) {
        aCallback(aDataStream, aStatusCode);

        HiddenCanvas.destroy(canvas);
        HiddenFrame.destroy(aFrame);
      });
    });

    // creating the hidden frame takes a while so let's create the canvas in
    // the meantime while we're waiting for the hidden frame to load
    let canvas = HiddenCanvas.create(aWidth, aHeight);
    let ctx = canvas.getContext("2d");

    // scale canvas accordingly
    let scaler = aWidth / frameWidth;
    ctx.scale(scaler, scaler);
  },

  _readCanvasData: function TN__readCanvasData(aCanvas, aCallback) {
    let dataUri = aCanvas.toDataURL(CONTENT_TYPE, "");
    let uri = Services.io.newURI(dataUri, "UTF8", null);

    gNetUtil.asyncFetch(uri, function (aDataStream, aStatus, aRequest) {
      aCallback(aDataStream);
    });
  }
};

// TODO
let HiddenFrame = {
  _hostFrameCallbacks: [],

  create: function HF_create(aUri, aWidth, aHeight, aCallback) {
    let self = this;

    this._getHostFrame(function (aFrame) {
      let doc = aFrame.contentDocument;
      let iframe = doc.createElementNS(XUL_NS, "iframe");
      iframe.setAttribute("type", "content");
      doc.documentElement.appendChild(iframe);

      iframe.style.width = aWidth + "px";
      iframe.style.height = aHeight + "px";

      // hide the scrollbars
      iframe.style.overflow = "hidden";

      // TODO
      let docShell = iframe.docShell;
      docShell.allowAuth = false;
      docShell.allowWindowControl = false;

      self._whenDocumentLoaded(iframe, function () {
        let statusCode = 200;

        try {
          // if the channel is a nsIHttpChannel get its http status
          let channel = docShell.currentDocumentChannel;
          let httpChannel = channel.QueryInterface(Ci.nsIHttpChannel);
          statusCode = httpChannel.responseStatus;
        } catch (e) {
          // not a http channel, ignore
        }

        aCallback(iframe, statusCode);
      });

      iframe.setAttribute("src", aUri);
    });
  },

  destroy: function HF_destroy(aFrame) {
    aFrame.parentNode.removeChild(aFrame);
  },

  _getHostFrame: function HF__getHostFrame(aCallback) {
    // return the cached value if any
    if (this._hostFrame) {
      aCallback(this._hostFrame);
      return;
    }

    if (this._hostFrameCallbacks.push(aCallback) > 1)
      return;

    // create the host frame
    let window = gAppShellService.hiddenDOMWindow;
    let doc = window.document;

    let hostFrame = doc.createElement("iframe");
    doc.body.appendChild(hostFrame);

    // TODO
    hostFrame.setAttribute("src", XUL_FRAME_SRC);

    let self = this;

    this._whenContentLoaded(hostFrame, function () {
      self._hostFrame = hostFrame;
      self._hostFrameCallbacks.forEach(function (callback) callback(hostFrame));
      self._hostFrameCallbacks = null;
    });
  },

  _whenContentLoaded: function HF__whenContentLoaded(aElement, aCallback) {
    aElement.addEventListener("DOMContentLoaded", function onLoad() {
      aElement.removeEventListener("DOMContentLoaded", onLoad, false);
      aCallback();
    }, false);
  },

  _whenDocumentLoaded: function HF__whenDocumentLoaded(aIframe, aCallback) {
    aIframe.addEventListener("load", function onLoad(aEvent) {
      if (aIframe.contentDocument == aEvent.target) {
        aIframe.removeEventListener("load", onLoad, true);
        aCallback();
      }
    }, true);
  }
};

// TODO
let HiddenCanvas = {
  create: function HC_create(aWidth, aHeight) {
    let window = gAppShellService.hiddenDOMWindow;
    let doc = window.document;

    let canvas = doc.createElement("canvas");
    doc.body.appendChild(canvas);

    canvas.setAttribute("moz-opaque", "");
    canvas.style.width = aWidth + "px";
    canvas.style.height = aHeight + "px";
    canvas.width = aWidth;
    canvas.height = aHeight;

    return canvas;
  },

  destroy: function HC_destroy(aCanvas) {
    aCanvas.parentNode.removeChild(aCanvas);
  }
};
